#
# Copyright (c) 2015, Prometheus Research, LLC
#


import colander

from six import iteritems

from .common import ValidationError, sub_schema, LanguageTag, \
    LocalizedMapping, IdentifierString, Options, LocalizedString, \
    Descriptor as BaseDescriptor, LocalizationChecker, \
    validate_instrument_version
from .instrument import InstrumentReference, get_full_type_definition


__all__ = (
    'ELEMENT_TYPES_ALL',
    'EVENT_ACTIONS_ALL',
    'UNPROMPTED_ACTIONS_ALL',
    'PARAMETER_TYPES_ALL',

    'Descriptor',
    'DescriptorList',
    'UrlList',
    'AudioSource',
    'TagList',
    'ElementType',
    'TextElementOptions',
    'AudioElementOptions',
    'Options',
    'Widget',
    'Expression',
    'EventAction',
    'EventTargetList',
    'EventList',
    'QuestionList',
    'QuestionElementOptions',
    'Element',
    'ElementList',
    'Page',
    'PageList',
    'UnpromptedAction',
    'CalculateUnpromptedOptions',
    'Unprompted',
    'UnpromptedCollection',
    'ParameterType',
    'ParameterCollection',
    'Form',
)


ELEMENT_TYPES_ALL = (
    'question',
    'header',
    'text',
    'divider',
    'audio',
)


EVENT_ACTIONS_ALL = (
    'hide',
    'disable',
    'hideEnumeration',
    'fail',
    'calculate',
)


UNPROMPTED_ACTIONS_ALL = (
    'calculate',
)


PARAMETER_TYPES_ALL = (
    'text',
    'numeric',
    'boolean',
)


# pylint: disable=abstract-method


class UrlList(colander.SequenceSchema):
    url = colander.SchemaNode(colander.String())
    validator = colander.Length(min=1)


class AudioSource(LocalizedMapping):
    def __init__(self, *args, **kwargs):
        super(AudioSource, self).__init__(
            UrlList(),
            *args,
            **kwargs
        )


class TagList(colander.SequenceSchema):
    tag = IdentifierString()
    validator = colander.Length(min=1)


class ElementType(colander.SchemaNode):
    schema_type = colander.String
    validator = colander.OneOf(ELEMENT_TYPES_ALL)


class TextElementOptions(colander.SchemaNode):
    text = LocalizedString()

    def __init__(self, *args, **kwargs):
        kwargs['typ'] = colander.Mapping(unknown='raise')
        super(TextElementOptions, self).__init__(*args, **kwargs)


class AudioElementOptions(colander.SchemaNode):
    source = AudioSource()

    def __init__(self, *args, **kwargs):
        kwargs['typ'] = colander.Mapping(unknown='raise')
        super(AudioElementOptions, self).__init__(*args, **kwargs)


class Widget(colander.SchemaNode):
    type = colander.SchemaNode(colander.String())
    options = Options(missing=colander.drop)

    def __init__(self, *args, **kwargs):
        kwargs['typ'] = colander.Mapping(unknown='raise')
        super(Widget, self).__init__(*args, **kwargs)


class Expression(colander.SchemaNode):
    schema_type = colander.String


class EventAction(colander.SchemaNode):
    schema_type = colander.String
    validator = colander.OneOf(EVENT_ACTIONS_ALL)


class EventTargetList(colander.SequenceSchema):
    target = IdentifierString()
    validator = colander.Length(min=1)


class Event(colander.SchemaNode):
    trigger = Expression()
    action = EventAction()
    targets = EventTargetList(missing=colander.drop)
    options = Options(missing=colander.drop)

    def __init__(self, *args, **kwargs):
        kwargs['typ'] = colander.Mapping(unknown='raise')
        super(Event, self).__init__(*args, **kwargs)


class EventList(colander.SequenceSchema):
    event = Event()
    validator = colander.Length(min=1)


class QuestionList(colander.SchemaNode):
    validator = colander.Length(min=1)

    def __init__(self, *args, **kwargs):
        kwargs['typ'] = colander.Sequence()
        super(QuestionList, self).__init__(*args, **kwargs)
        self.add(QuestionElementOptions(
            allow_complex=False,
            name='question',
        ))


class Descriptor(BaseDescriptor):
    audio = AudioSource(missing=colander.drop)


class DescriptorList(colander.SequenceSchema):
    descriptor = Descriptor()
    validator = colander.Length(min=1)


class QuestionElementOptions(colander.SchemaNode):
    fieldId = IdentifierString()
    text = LocalizedString()
    audio = AudioSource(missing=colander.drop)
    help = LocalizedString(missing=colander.drop)
    error = LocalizedString(missing=colander.drop)
    enumerations = DescriptorList(missing=colander.drop)
    widget = Widget(missing=colander.drop)
    events = EventList(missing=colander.drop)

    def __init__(self, *args, **kwargs):
        self.allow_complex = kwargs.pop('allow_complex', True)
        kwargs['typ'] = colander.Mapping(unknown='raise')
        super(QuestionElementOptions, self).__init__(*args, **kwargs)
        if self.allow_complex:
            self.add(QuestionList(
                name='questions',
                missing=colander.drop,
            ))
            self.add(DescriptorList(
                name='rows',
                missing=colander.drop,
            ))


ELEMENT_TYPE_OPTION_VALIDATORS = {
    'question': QuestionElementOptions(),
    'text': TextElementOptions(),
    'header': TextElementOptions(),
    'audio': AudioElementOptions(),
}


class Element(colander.SchemaNode):
    type = ElementType()
    options = Options(missing=colander.drop)
    tags = TagList(missing=colander.drop)

    def __init__(self, *args, **kwargs):
        kwargs['typ'] = colander.Mapping(unknown='raise')
        super(Element, self).__init__(*args, **kwargs)

    def validator(self, node, cstruct):
        element_type = cstruct.get('type', None)
        validator = ELEMENT_TYPE_OPTION_VALIDATORS.get(element_type, None)
        options = cstruct.get('options', None)
        if validator:
            sub_schema(
                validator,
                node.get('options'),
                options,
            )
        elif options is not None:
            raise ValidationError(
                node.get('options'),
                '"%s" elements do not accept options' % element_type,
            )


class ElementList(colander.SequenceSchema):
    element = Element()
    validator = colander.Length(min=1)


class Page(colander.SchemaNode):
    id = IdentifierString()  # pylint: disable=invalid-name
    elements = ElementList()

    def __init__(self, *args, **kwargs):
        kwargs['typ'] = colander.Mapping(unknown='raise')
        super(Page, self).__init__(*args, **kwargs)


class PageList(colander.SequenceSchema):
    page = Page()

    def validator(self, node, cstruct):
        if len(cstruct) < 1:
            raise ValidationError(
                node,
                'Shorter than minimum length 1',
            )

        ids = [page['id'] for page in cstruct]
        if len(ids) != len(set(ids)):
            raise ValidationError(
                node,
                'Page IDs must be unique',
            )


class UnpromptedAction(colander.SchemaNode):
    schema_type = colander.String
    validator = colander.OneOf(UNPROMPTED_ACTIONS_ALL)


class CalculateUnpromptedOptions(colander.SchemaNode):
    calculation = Expression()

    def __init__(self, *args, **kwargs):
        kwargs['typ'] = colander.Mapping(unknown='raise')
        super(CalculateUnpromptedOptions, self).__init__(*args, **kwargs)


UNPROMPTED_ACTION_OPTION_VALIDATORS = {
    'calculate': CalculateUnpromptedOptions(),
}


class Unprompted(colander.SchemaNode):
    action = UnpromptedAction()
    options = Options(missing=colander.drop)

    def __init__(self, *args, **kwargs):
        kwargs['typ'] = colander.Mapping(unknown='raise')
        super(Unprompted, self).__init__(*args, **kwargs)

    def validator(self, node, cstruct):
        action = cstruct.get('action', None)
        validator = UNPROMPTED_ACTION_OPTION_VALIDATORS.get(action, None)
        options = cstruct.get('options', None)
        if validator:
            sub_schema(
                validator,
                node.get('options'),
                options,
            )
        elif options is not None:
            raise ValidationError(
                node.get('options'),
                '"%s" actions do not accept options' % action,
            )


class UnpromptedCollection(colander.SchemaNode):
    def __init__(self, *args, **kwargs):
        kwargs['typ'] = colander.Mapping(unknown='preserve')
        super(UnpromptedCollection, self).__init__(*args, **kwargs)

    def validator(self, node, cstruct):
        cstruct = cstruct or {}
        if len(cstruct) == 0:
            raise ValidationError(
                node,
                'At least one key/value pair must be defined',
            )

        for name, defn in iteritems(cstruct):
            sub_schema(IdentifierString, node, name)
            sub_schema(Unprompted, node, defn)


class ParameterType(colander.SchemaNode):
    schema_type = colander.String
    validator = colander.OneOf(PARAMETER_TYPES_ALL)


class ParameterOptions(colander.SchemaNode):
    type = ParameterType()

    def __init__(self, *args, **kwargs):
        kwargs['typ'] = colander.Mapping(unknown='raise')
        super(ParameterOptions, self).__init__(*args, **kwargs)


class ParameterCollection(colander.SchemaNode):
    def __init__(self, *args, **kwargs):
        kwargs['typ'] = colander.Mapping(unknown='preserve')
        super(ParameterCollection, self).__init__(*args, **kwargs)

    def validator(self, node, cstruct):
        cstruct = cstruct or {}
        if len(cstruct) == 0:
            raise ValidationError(
                node,
                'At least one key/value pair must be defined',
            )

        for name, options in iteritems(cstruct):
            sub_schema(IdentifierString, node, name)
            sub_schema(ParameterOptions, node, options)


class Form(colander.SchemaNode):
    instrument = InstrumentReference()
    defaultLocalization = LanguageTag()
    title = LocalizedString(missing=colander.drop)
    pages = PageList()
    unprompted = UnpromptedCollection(missing=colander.drop)
    parameters = ParameterCollection(missing=colander.drop)

    def __init__(self, instrument=None, *args, **kwargs):
        self.instrument = instrument
        kwargs['typ'] = colander.Mapping(unknown='raise')
        super(Form, self).__init__(*args, **kwargs)

    def validator(self, node, cstruct):
        self._check_localizations(node, cstruct)

        if not self.instrument:
            return

        validate_instrument_version(
            self.instrument,
            cstruct,
            node.get('instrument'),
        )

        self._check_fields_covered(node, cstruct)

        self._check_type_specifics(node, cstruct)

    def _check_localizations(self, node, cstruct):
        checker = LocalizationChecker(node, cstruct['defaultLocalization'])

        def _ensure_element(element):
            if 'options' not in element:
                return
            options = element['options']

            checker.ensure(options, 'text', scope='Element Text')
            checker.ensure(options, 'help', scope='Element Help')
            checker.ensure(options, 'error', scope='Element Error')
            checker.ensure(options, 'audio', scope='Element Audio')
            checker.ensure(options, 'source', scope='Audio Source')

            for question in options.get('questions', []):
                _ensure_element(question)
            for enumeration in options.get('enumerations', []):
                checker.ensure_descriptor(enumeration, scope='Enumeration')
            for row in options.get('rows', []):
                checker.ensure_descriptor(row, scope='Matrix Row')

        checker.ensure(cstruct, 'title', node=node.get('title'))
        for page in cstruct['pages']:
            for element in page['elements']:
                _ensure_element(element)

    def _check_fields_covered(self, node, cstruct):
        instrument_fields = set([
            field['id']
            for field in self.instrument['record']
        ])

        form_fields = set()
        for page in cstruct['pages']:
            for element in page['elements']:
                if element['type'] != 'question':
                    continue

                field_id = element['options']['fieldId']
                if field_id in form_fields:
                    raise ValidationError(
                        node,
                        'Field "%s" is addressed by more than one question' % (
                            field_id,
                        )
                    )
                else:
                    form_fields.add(field_id)

        missing = instrument_fields - form_fields
        if missing:
            raise ValidationError(
                node,
                'There are Instrument fields which are missing: %s' % (
                    ', '.join(missing),
                )
            )

        extra = form_fields - instrument_fields
        if extra:
            raise ValidationError(
                node,
                'There are extra fields referenced by questions: %s' % (
                    ', '.join(extra),
                )
            )

    def _get_instrument_field(self, name):
        for field in self.instrument['record']:
            if field['id'] == name:
                return field

    def _check_type_specifics(self, node, cstruct):
        for page in cstruct['pages']:
            for element in page['elements']:
                if element['type'] != 'question':
                    continue

                type_def = get_full_type_definition(
                    self.instrument,
                    self._get_instrument_field(
                        element['options']['fieldId'],
                    )['type'],
                )

                if type_def['base'] not in ('enumeration', 'enumerationSet') \
                        and 'enumerations' in element['options']:
                    raise ValidationError(
                        node,
                        'Field "%s" cannot have an enumerations'
                        ' configuration' % (
                            element['options']['fieldId'],
                        ),
                    )

                self._check_matrix(node, type_def, element)
                self._check_subquestions(node, type_def, element)

    def _check_matrix(self, node, type_def, element):
        if type_def['base'] == 'matrix':
            instrument_rows = set([
                row['id']
                for row in type_def['rows']
            ])

            form_rows = set()
            for row in element['options'].get('rows', []):
                if row['id'] in form_rows:
                    raise ValidationError(
                        node,
                        'Row %s is addressed by more than one descriptor in'
                        ' %s' % (
                            row['id'],
                            element['options']['fieldId'],
                        ),
                    )
                else:
                    form_rows.add(row['id'])

            missing = instrument_rows - form_rows
            if missing:
                raise ValidationError(
                    node,
                    'There are missing rows in %s: %s' % (
                        element['options']['fieldId'],
                        ', '.join(missing),
                    )
                )

            extra = form_rows - instrument_rows
            if extra:
                raise ValidationError(
                    node,
                    'There are extra rows referenced by %s: %s' % (
                        element['options']['fieldId'],
                        ', '.join(extra),
                    )
                )

        elif 'rows' in element['options']:
            raise ValidationError(
                node,
                'Field "%s" cannot have a rows configuration' % (
                    element['options']['fieldId'],
                ),
            )

    def _check_subquestions(self, node, type_def, element):
        if type_def['base'] in ('matrix', 'recordList'):
            instrument_fields = set([
                field['id']
                for field in type_def[
                    'columns' if type_def['base'] == 'matrix' else 'record'
                ]
            ])

            form_fields = set()
            for subfield in element['options'].get('questions', []):
                if subfield['fieldId'] in form_fields:
                    raise ValidationError(
                        node,
                        'Subfield %s is addressed by more than one question in'
                        ' %s' % (
                            subfield['fieldId'],
                            element['options']['fieldId'],
                        ),
                    )
                else:
                    form_fields.add(subfield['fieldId'])

            missing = instrument_fields - form_fields
            if missing:
                raise ValidationError(
                    node,
                    'There are missing subfields in %s: %s' % (
                        element['options']['fieldId'],
                        ', '.join(missing),
                    )
                )

            extra = form_fields - instrument_fields
            if extra:
                raise ValidationError(
                    node,
                    'There are extra subfields referenced by %s: %s' % (
                        element['options']['fieldId'],
                        ', '.join(extra),
                    )
                )

        elif 'questions' in element['options']:
            raise ValidationError(
                node,
                'Field "%s" cannot have a questions configuration' % (
                    element['options']['fieldId'],
                ),
            )




#
# Copyright (c) 2015, Prometheus Research, LLC
#


from .common import SortedDict, TypedDefinedOrderDict, InstrumentReference


__all__ = (
    'CalculationSet',
)


class Calculation(TypedDefinedOrderDict):
    order = [
        'id',
        'description',
        'type',
        'method',
        'options',
    ]

    key_types = {
        'options': SortedDict,
    }


class CalculationSet(TypedDefinedOrderDict):
    order = [
        'instrument',
        'calculations',
    ]

    key_types = {
        'instrument': InstrumentReference,
        'calculations': [Calculation],
    }




#
# Copyright (c) 2015, Prometheus Research, LLC
#


import colander

from .common import ValidationError, sub_schema, Options, \
    validate_instrument_version
from .instrument import InstrumentReference, IdentifierString, Description, \
    TYPES_SIMPLE


__all__ = (
    'METHODS_ALL',

    'CalculationResultType',
    'CalculationMethod',
    'Expression',
    'PythonOptions',
    'HtsqlOptions',
    'Calculation',
    'CalculationList',
    'CalculationSet',
)


METHODS_ALL = (
    'python',
    'htsql',
)


# pylint: disable=abstract-method


class CalculationResultType(colander.SchemaNode):
    schema_type = colander.String
    validator = colander.OneOf(TYPES_SIMPLE)


class CalculationMethod(colander.SchemaNode):
    schema_type = colander.String
    validator = colander.OneOf(METHODS_ALL)


class Expression(colander.SchemaNode):
    schema_type = colander.String


class PythonOptions(colander.SchemaNode):
    expression = Expression(missing=colander.drop)
    callable = Expression(missing=colander.drop)

    def __init__(self, *args, **kwargs):
        kwargs['typ'] = colander.Mapping(unknown='raise')
        super(PythonOptions, self).__init__(*args, **kwargs)

    def validator(self, node, cstruct):
        if ('expression' in cstruct) == ('callable' in cstruct):
            raise ValidationError(
                node,
                'Exactly one option of "exression" or "callable" must be'
                ' specified',
            )


class HtsqlOptions(colander.SchemaNode):
    expression = Expression()

    def __init__(self, *args, **kwargs):
        kwargs['typ'] = colander.Mapping(unknown='raise')
        super(HtsqlOptions, self).__init__(*args, **kwargs)


METHOD_OPTION_VALIDATORS = {
    'python': PythonOptions(),
    'htsql': HtsqlOptions(),
}


class Calculation(colander.SchemaNode):
    id = IdentifierString()  # pylint: disable=invalid-name
    description = Description()
    type = CalculationResultType()
    method = CalculationMethod()
    options = Options(missing=colander.drop)

    def __init__(self, *args, **kwargs):
        kwargs['typ'] = colander.Mapping(unknown='raise')
        super(Calculation, self).__init__(*args, **kwargs)

    def validator(self, node, cstruct):
        method = cstruct.get('method', None)
        validator = METHOD_OPTION_VALIDATORS.get(method, None)
        options = cstruct.get('options', None)
        if validator:
            sub_schema(
                validator,
                node.get('options'),
                options,
            )
        elif options is not None:
            raise ValidationError(
                node.get('options'),
                'The "%s" method does not accept options' % method,
            )


class CalculationList(colander.SequenceSchema):
    calculation = Calculation()

    def validator(self, node, cstruct):
        if len(cstruct) < 1:
            raise ValidationError(
                node,
                'Shorter than minimum length 1',
            )

        ids = [calculation['id'] for calculation in cstruct]
        if len(ids) != len(set(ids)):
            raise ValidationError(
                node,
                'Calculation IDs must be unique',
            )


class CalculationSet(colander.SchemaNode):
    instrument = InstrumentReference()
    calculations = CalculationList()

    def __init__(self, instrument=None, *args, **kwargs):
        self.instrument = instrument
        kwargs['typ'] = colander.Mapping(unknown='raise')
        super(CalculationSet, self).__init__(*args, **kwargs)

    def validator(self, node, cstruct):
        if not self.instrument:
            return

        validate_instrument_version(
            self.instrument,
            cstruct,
            node.get('instrument'),
        )




#
# Copyright (c) 2015, Prometheus Research, LLC
#


from .common import SortedDict, TypedDefinedOrderDict, InstrumentReference, \
    Descriptor


__all__ = (
    'Interaction',
)


class StepOptions(TypedDefinedOrderDict):
    order = [
        'fieldId',
        'text',
        'error',
        'enumerations',
    ]

    key_types = {
        'text': SortedDict,
        'error': SortedDict,
        'enumerations': [Descriptor],
    }


class Step(TypedDefinedOrderDict):
    order = [
        'type',
        'options',
    ]

    key_types = {
        'options': StepOptions,
    }


class TimeoutDetails(TypedDefinedOrderDict):
    order = [
        'threshold',
        'text',
    ]

    key_types = {
        'text': SortedDict,
    }


class Timeout(TypedDefinedOrderDict):
    order = [
        'warn',
        'abort',
    ]

    key_types = {
        'warn': TimeoutDetails,
        'abort': TimeoutDetails,
    }


class Interaction(TypedDefinedOrderDict):
    order = [
        'instrument',
        'defaultLocalization',
        'defaultTimeout',
        'steps',
    ]

    key_types = {
        'instrument': InstrumentReference,
        'defaultTimeout': Timeout,
        'steps': [Step],
    }




#
# Copyright (c) 2015, Prometheus Research, LLC
#


import colander

from .common import ValidationError, sub_schema, LanguageTag, \
    IdentifierString, Options, LocalizedString, DescriptorList, \
    LocalizationChecker, validate_instrument_version
from .instrument import InstrumentReference, TYPES_COMPLEX, \
    get_full_type_definition


__all__ = (
    'Interaction',
)


STEP_TYPES_ALL = (
    'question',
    'text',
)


# pylint: disable=abstract-method


class StepType(colander.SchemaNode):
    schema_type = colander.String
    validator = colander.OneOf(STEP_TYPES_ALL)


class TextStepOptions(colander.SchemaNode):
    text = LocalizedString()

    def __init__(self, *args, **kwargs):
        kwargs['typ'] = colander.Mapping(unknown='raise')
        super(TextStepOptions, self).__init__(*args, **kwargs)


class QuestionStepOptions(colander.SchemaNode):
    fieldId = IdentifierString()
    text = LocalizedString()
    error = LocalizedString(missing=colander.drop)
    enumerations = DescriptorList(missing=colander.drop)

    def __init__(self, *args, **kwargs):
        kwargs['typ'] = colander.Mapping(unknown='raise')
        super(QuestionStepOptions, self).__init__(*args, **kwargs)


STEP_TYPE_OPTION_VALIDATORS = {
    'question': QuestionStepOptions(),
    'text': TextStepOptions(),
}


class Step(colander.SchemaNode):
    type = StepType()
    options = Options(missing=colander.drop)

    def __init__(self, *args, **kwargs):
        kwargs['typ'] = colander.Mapping(unknown='raise')
        super(Step, self).__init__(*args, **kwargs)

    def validator(self, node, cstruct):
        step_type = cstruct.get('type', None)
        validator = STEP_TYPE_OPTION_VALIDATORS.get(step_type, None)
        options = cstruct.get('options', None)
        if validator:
            sub_schema(
                validator,
                node.get('options'),
                options,
            )
        elif options is not None:
            raise ValidationError(
                node.get('options'),
                '"%s" step do not accept options' % step_type,
            )


class StepList(colander.SequenceSchema):
    step = Step()
    validator = colander.Length(min=1)


class Threshold(colander.SchemaNode):
    schema_type = colander.Integer
    validator = colander.Range(min=1)


class TimeoutDetails(colander.SchemaNode):
    threshold = Threshold()
    text = LocalizedString()

    def __init__(self, *args, **kwargs):
        kwargs['typ'] = colander.Mapping(unknown='raise')
        super(TimeoutDetails, self).__init__(*args, **kwargs)


class Timeout(colander.SchemaNode):
    warn = TimeoutDetails(missing=colander.drop)
    abort = TimeoutDetails(missing=colander.drop)

    def __init__(self, *args, **kwargs):
        kwargs['typ'] = colander.Mapping(unknown='raise')
        super(Timeout, self).__init__(*args, **kwargs)

    def validator(self, node, cstruct):
        if not cstruct.get('warn') and not cstruct.get('abort'):
            raise ValidationError(
                node,
                'At least one of "warn" or "abort" must be defined',
            )


class Interaction(colander.SchemaNode):
    instrument = InstrumentReference()
    defaultLocalization = LanguageTag()
    defaultTimeout = Timeout(missing=colander.drop)
    steps = StepList()

    def __init__(self, instrument=None, *args, **kwargs):
        self.instrument = instrument
        kwargs['typ'] = colander.Mapping(unknown='raise')
        super(Interaction, self).__init__(*args, **kwargs)

    def validator(self, node, cstruct):
        self._check_localizations(node, cstruct)

        if not self.instrument:
            return

        validate_instrument_version(
            self.instrument,
            cstruct,
            node.get('instrument'),
        )

        self._check_fields_covered(node, cstruct)

        self._check_type_specifics(node, cstruct)

    def _check_localizations(self, node, cstruct):
        checker = LocalizationChecker(node, cstruct['defaultLocalization'])

        timeouts = cstruct.get('defaultTimeout', {})
        for level in ('warn', 'abort'):
            if level in timeouts:
                checker.ensure(
                    timeouts[level],
                    'text',
                    scope='Timeout %s Text' % level,
                )

        for step in cstruct['steps']:
            if 'options' not in step:  # pragma: no cover
                return
            options = step['options']

            checker.ensure(options, 'text', scope='Step Text')
            checker.ensure(options, 'error', scope='Step Error')

            for enumeration in options.get('enumerations', []):
                checker.ensure_descriptor(enumeration, scope='Enumeration')

    def _check_fields_covered(self, node, cstruct):
        instrument_fields = set([
            field['id']
            for field in self.instrument['record']
        ])

        intr_fields = set()
        for step in cstruct['steps']:
            if step['type'] != 'question':
                continue

            field_id = step['options']['fieldId']
            if field_id in intr_fields:
                raise ValidationError(
                    node.get('steps'),
                    'Field "%s" is addressed by more than one question' % (
                        field_id,
                    )
                )
            else:
                intr_fields.add(field_id)

        missing = instrument_fields - intr_fields
        if missing:
            raise ValidationError(
                node.get('steps'),
                'There are Instrument fields which are missing: %s' % (
                    ', '.join(missing),
                )
            )

        extra = intr_fields - instrument_fields
        if extra:
            raise ValidationError(
                node.get('steps'),
                'There are extra fields referenced by questions: %s' % (
                    ', '.join(extra),
                )
            )

    def _get_instrument_field(self, name):
        for field in self.instrument['record']:
            if field['id'] == name:
                return field

    def _check_type_specifics(self, node, cstruct):
        for step in cstruct['steps']:
            if step['type'] != 'question':
                continue

            type_def = get_full_type_definition(
                self.instrument,
                self._get_instrument_field(
                    step['options']['fieldId'],
                )['type'],
            )

            if type_def['base'] in TYPES_COMPLEX:
                raise ValidationError(
                    node.get('steps'),
                    'Complex Instrument Types are not allowed in Interactions',
                )

            if type_def['base'] not in ('enumeration', 'enumerationSet') \
                    and 'enumerations' in step['options']:
                raise ValidationError(
                    node.get('steps'),
                    'Field "%s" cannot have an enumerations configuration' % (
                        step['options']['fieldId'],
                    ),
                )




#
# Copyright (c) 2015, Prometheus Research, LLC
#


import json

import six

from .common import ValidationError
from .assessment import Assessment
from .calculationset import CalculationSet
from .form import Form
from .instrument import Instrument
from .interaction import Interaction


__all__ = (
    'ValidationError',
    'validate_instrument',
    'validate_assessment',
    'validate_calculationset',
    'validate_form',
    'validate_interaction',
)


def _get_struct(src):
    if isinstance(src, six.string_types):
        src = json.loads(src)
    elif hasattr(src, 'read'):
        src = json.load(src)
    return src


def validate_instrument(instrument):
    """
    Validates the input against the PRISMH Instrument Definition
    specification.

    :param instrument: The Instrument Definition to validate
    :type instrument: JSON string, dict, or file-like object
    :raises ValidationError: If the input fails any part of the specification
    """

    instrument = _get_struct(instrument)
    validator = Instrument()
    validator.deserialize(instrument)


def validate_assessment(assessment, instrument=None):
    """
    Validates the input against the PRISMH Assessment Document specification.

    :param assessment: The Assessment Document to validate
    :type assessment: JSON string, dict, or file-like object
    :param instrument:
        The Instrument Definition to validate the Assessment against. If not
        specified, this defaults to ``None``, which means that only the basic
        structure of the Assessment will be validated -- not its conformance to
        the Instrument.
    :type instrument: JSON string, dict, or file-like object
    :raises ValidationError: If the input fails any part of the specification
    """

    assessment = _get_struct(assessment)
    if instrument:
        instrument = _get_struct(instrument)
        validate_instrument(instrument)
    validator = Assessment(instrument=instrument)
    validator.deserialize(assessment)


def validate_calculationset(calculationset, instrument=None):
    """
    Validates the input against the PRISMH Calculation Set Definition
    specification.

    :param form: The Calculation Set Definition to validate
    :type form: JSON string, dict, or file-like object
    :param instrument:
        The Instrument Definition to validate the Calculation Set against. If
        not specified, this defaults to ``None``, which means that only the
        basic structure of the Calculation Set will be validated -- not its
        conformance to the Instrument.
    :type instrument: JSON string, dict, or file-like object
    :raises ValidationError: If the input fails any part of the specification
    """

    calculationset = _get_struct(calculationset)
    if instrument:
        instrument = _get_struct(instrument)
        validate_instrument(instrument)
    validator = CalculationSet(instrument=instrument)
    validator.deserialize(calculationset)


def validate_form(form, instrument=None):
    """
    Validates the input against the PRISMH Web Form Configuration
    specification.

    :param form: The Web Form Configuration to validate
    :type form: JSON string, dict, or file-like object
    :param instrument:
        The Instrument Definition to validate the Form against. If not
        specified, this defaults to ``None``, which means that only the basic
        structure of the Form will be validated -- not its conformance to
        the Instrument.
    :type instrument: JSON string, dict, or file-like object
    :raises ValidationError: If the input fails any part of the specification
    """

    form = _get_struct(form)
    if instrument:
        instrument = _get_struct(instrument)
        validate_instrument(instrument)
    validator = Form(instrument=instrument)
    validator.deserialize(form)


def validate_interaction(interaction, instrument=None):
    """
    Validates the input against the PRISMH SMS Interaction Configuration
    specification.

    :param interaction: The SMS Interaction Configuration to validate
    :type interaction: JSON string, dict, or file-like object
    :param instrument:
        The Instrument Definition to validate the Interaction against. If not
        specified, this defaults to ``None``, which means that only the basic
        structure of the Interaction will be validated -- not its conformance
        to the Instrument.
    :type instrument: JSON string, dict, or file-like object
    :raises ValidationError: If the input fails any part of the specification
    """

    interaction = _get_struct(interaction)
    if instrument:
        instrument = _get_struct(instrument)
        validate_instrument(instrument)
    validator = Interaction(instrument=instrument)
    validator.deserialize(interaction)




#
# Copyright (c) 2015, Prometheus Research, LLC
#


from six import string_types

from .common import SortedDict, TypedSortedDict, TypedDefinedOrderDict, \
    InstrumentReference


__all__ = (
    'Assessment',
)


class Value(TypedDefinedOrderDict):
    order = [
        'value',
        'annotation',
        'explanation',
        'meta',
    ]

    key_types = {
        'meta': SortedDict,
    }

    def __init__(self, value=None):
        super(Value, self).__init__(value or {})
        if 'value' in self:
            self['value'] = self['value']

    def __setitem__(self, key, value):
        if key == 'value':
            if isinstance(value, dict):
                value = ValueCollectionMapping(value)
            elif isinstance(value, list):
                if not isinstance(value[0], string_types):
                    value = [
                        ValueCollection(v)
                        for v in value
                    ]
        super(Value, self).__setitem__(key, value)


class ValueCollection(TypedSortedDict):
    subtype = Value


class ValueCollectionMapping(TypedSortedDict):
    subtype = ValueCollection


class Assessment(TypedDefinedOrderDict):
    order = [
        'instrument',
        'meta',
        'values',
    ]

    key_types = {
        'instrument': InstrumentReference,
        'meta': SortedDict,
        'values': ValueCollection
    }




#
# Copyright (c) 2015, Prometheus Research, LLC
#


from .common import SortedDict, TypedSortedDict, DefinedOrderDict, \
    TypedDefinedOrderDict


__all__ = (
    'MatrixRow',
    'BoundConstraint',
    'InstrumentField',
    'InstrumentType',
    'InstrumentTypeCollection',
    'Instrument',
)


class MatrixRow(DefinedOrderDict):
    order = [
        'id',
        'description',
        'required',
    ]


class BoundConstraint(DefinedOrderDict):
    order = [
        'min',
        'max',
    ]


class InstrumentField(DefinedOrderDict):
    order = [
        'id',
        'description',
        'type',
        'required',
        'identifiable',
        'annotation',
        'explanation',
    ]

    def __init__(self, field=None):
        super(InstrumentField, self).__init__(field or {})
        if 'type' in self:
            self['type'] = self['type']

    def __setitem__(self, key, value):
        if key == 'type' and isinstance(value, dict):
            value = InstrumentType(value)
        super(InstrumentField, self).__setitem__(key, value)


class InstrumentType(TypedDefinedOrderDict):
    order = [
        'base',
        'range',
        'length',
        'pattern',
        'enumerations',
        'record',
        'columns',
        'rows',
    ]

    key_types = {
        'record': [InstrumentField],
        'columns': [InstrumentField],
        'length': BoundConstraint,
        'range': BoundConstraint,
        'rows': [MatrixRow],
        'enumerations': SortedDict,
    }


class InstrumentTypeCollection(TypedSortedDict):
    subtype = InstrumentType


class Instrument(TypedDefinedOrderDict):
    order = [
        'id',
        'version',
        'title',
        'description',
        'types',
        'record',
    ]

    key_types = {
        'version': str,
        'types': InstrumentTypeCollection,
        'record': [InstrumentField],
    }




#
# Copyright (c) 2015, Prometheus Research, LLC
#


import re

from collections import Callable

import colander

from six import iteritems


__all__ = (
    'RE_IDENTIFIER',

    'ValidationError',

    'IdentifierString',
    'AnyType',
    'OneOfType',
    'StrictBooleanType',
    'OptionalStringType',
    'LanguageTag',
    'LocalizedMapping',
    'LocalizedString',
    'Options',
    'Descriptor',
    'DescriptorList',

    'LocalizationChecker',
    'sub_schema',
    'validate_instrument_version',
)


# pylint: disable=abstract-method


ValidationError = colander.Invalid


RE_IDENTIFIER = re.compile(r'^[a-z](?:[a-z0-9]|[_](?![_]))*[a-z0-9]$')


class IdentifierString(colander.SchemaNode):
    schema_type = colander.String
    validator = colander.Regex(RE_IDENTIFIER)


def sub_schema(schema, node, cstruct):
    if not isinstance(schema, colander.SchemaNode):
        schema = schema()
    try:
        schema.deserialize(cstruct)
    except ValidationError as exc:
        exc.node = node
        raise exc


class AnyType(colander.SchemaType):
    # pylint: disable=unused-argument
    def deserialize(self, node, cstruct):
        if cstruct is colander.null:
            return colander.null
        return cstruct


class OneOfType(colander.SchemaType):
    def __init__(self, *args):
        self.possible_types = [
            arg() if isinstance(arg, Callable) else arg
            for arg in args
        ]

    def deserialize(self, node, cstruct):
        for i in range(len(self.possible_types)):
            try:
                return self.possible_types[i].deserialize(node, cstruct)
            except ValidationError:
                if i == (len(self.possible_types) - 1):
                    raise


class StrictBooleanType(colander.SchemaType):
    def deserialize(self, node, cstruct):
        if cstruct is colander.null:
            return colander.null

        if isinstance(cstruct, bool):
            return cstruct

        raise ValidationError(
            node,
            '"%r" is not a boolean value' % (cstruct,)
        )


class OptionalStringType(colander.String):
    def deserialize(self, node, cstruct):
        if cstruct is colander.null or cstruct is None:
            return colander.null

        try:
            result = cstruct
            if isinstance(result, (colander.text_type, bytes)):
                if self.encoding:
                    result = colander.text_(cstruct, self.encoding)
                else:
                    result = colander.text_type(cstruct)
            else:
                raise ValidationError(node)
        except Exception as exc:
            raise ValidationError(
                node,
                colander._(
                    '${val} is not a string: ${err}',
                    mapping={
                        'val': cstruct,
                        'err': exc
                    }
                )
            )

        return result


RE_LANGUAGE_TAG = re.compile(
    r'^(((([A-Za-z]{2,3}(-([A-Za-z]{3}(-[A-Za-z]{3}){0,2}))?)|[A-Za-z]{4}|[A-Za-z]{5,8})(-([A-Za-z]{4}))?(-([A-Za-z]{2}|[0-9]{3}))?(-([A-Za-z0-9]{5,8}|[0-9][A-Za-z0-9]{3}))*(-([0-9A-WY-Za-wy-z](-[A-Za-z0-9]{2,8})+))*(-(x(-[A-Za-z0-9]{1,8})+))?)|(x(-[A-Za-z0-9]{1,8})+)|((en-GB-oed|i-ami|i-bnn|i-default|i-enochian|i-hak|i-klingon|i-lux|i-mingo|i-navajo|i-pwn|i-tao|i-tay|i-tsu|sgn-BE-FR|sgn-BE-NL|sgn-CH-DE)|(art-lojban|cel-gaulish|no-bok|no-nyn|zh-guoyu|zh-hakka|zh-min|zh-min-nan|zh-xiang)))$'  # noqa
)


class LanguageTag(colander.SchemaNode):
    schema_type = colander.String
    validator = colander.Regex(RE_LANGUAGE_TAG)


class LocalizedMapping(colander.SchemaNode):
    def __init__(self, sub_type, *args, **kwargs):
        self.sub_type = sub_type
        kwargs['typ'] = colander.Mapping(unknown='preserve')
        super(LocalizedMapping, self).__init__(*args, **kwargs)

    def validator(self, node, cstruct):
        cstruct = cstruct or {}

        if len(cstruct) == 0:
            raise ValidationError(
                node,
                'At least one localization must be specified',
            )

        for language_tag, translation in iteritems(cstruct):
            sub_schema(LanguageTag, node, language_tag)
            sub_schema(self.sub_type, node, translation)


class LocalizedString(LocalizedMapping):
    def __init__(self, *args, **kwargs):
        super(LocalizedString, self).__init__(
            colander.SchemaNode(colander.String()),
            *args,
            **kwargs
        )


class Options(colander.SchemaNode):
    def __init__(self, *args, **kwargs):
        kwargs['typ'] = colander.Mapping(unknown='preserve')
        super(Options, self).__init__(*args, **kwargs)

    def validator(self, node, cstruct):
        cstruct = cstruct or {}
        if len(cstruct) == 0:
            raise ValidationError(
                node,
                'At least one key/value pair must be defined',
            )


class Descriptor(colander.SchemaNode):
    id = colander.SchemaNode(colander.String())  # pylint: disable=invalid-name
    text = LocalizedString()
    help = LocalizedString(missing=colander.drop)

    def __init__(self, *args, **kwargs):
        kwargs['typ'] = colander.Mapping(unknown='raise')
        super(Descriptor, self).__init__(*args, **kwargs)


class DescriptorList(colander.SequenceSchema):
    descriptor = Descriptor()
    validator = colander.Length(min=1)


class LocalizationChecker(object):
    def __init__(self, node, default_localization):
        self.node = node
        self.locale = default_localization

    def ensure(self, obj, key, node=None, scope=None):
        if not isinstance(obj, dict) or key not in obj:
            return
        if self.locale not in obj[key]:
            raise ValidationError(
                node or self.node,
                '%sMissing default localization' % (
                    ('%s ' % scope) if scope else '',
                )
            )

    def ensure_descriptor(self, descriptor, scope=None):
        scope = scope or ''
        self.ensure(descriptor, 'text', scope=(scope + ' Text').strip())
        self.ensure(descriptor, 'help', scope=(scope + ' Help').strip())
        self.ensure(descriptor, 'audio', scope=(scope + ' Audio').strip())


def validate_instrument_version(instrument, obj, node):
    if instrument['id'] != obj['instrument']['id'] or \
            instrument['version'] != obj['instrument']['version']:
        raise ValidationError(
            node,
            'Incorrect Instrument version referenced',
        )




#
# Copyright (c) 2015, Prometheus Research, LLC
#


import re

from copy import deepcopy

import colander

from six import iteritems, string_types, integer_types

from .common import ValidationError, sub_schema, AnyType, LanguageTag, \
    validate_instrument_version
from .instrument import InstrumentReference, IdentifierString, \
    get_full_type_definition


__all__ = (
    'METADATA_SCOPE_ASSESSMENT',
    'METADATA_SCOPE_VALUE',
    'METADATA_STANDARD_PROPERTIES',

    'MetadataCollection',
    'ValueCollection',
    'Assessment',
)


RE_PRODUCT_TOKEN = re.compile(r'^(.+)/(.+)$')
RE_DATE = re.compile(r'^\d{4}-\d{2}-\d{2}$')
RE_TIME = re.compile(r'^\d{2}:\d{2}:\d{2}$')
RE_DATETIME = re.compile(r'^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}$')

METADATA_SCOPE_ASSESSMENT = 'assessment'
METADATA_SCOPE_VALUE = 'value'

METADATA_STANDARD_PROPERTIES = {
    METADATA_SCOPE_ASSESSMENT: {
        'language': LanguageTag(),
        'application': colander.SchemaNode(
            colander.String(),
            validator=colander.Regex(RE_PRODUCT_TOKEN),
        ),
        'dateCompleted': colander.SchemaNode(
            colander.DateTime(),
        ),
        'timeTaken': colander.SchemaNode(
            colander.Integer(),
        ),
    },

    METADATA_SCOPE_VALUE: {
        'timeTaken': colander.SchemaNode(
            colander.Integer(),
        ),
    },
}


# pylint: disable=abstract-method


class MetadataCollection(colander.SchemaNode):
    def __init__(self, scope, *args, **kwargs):
        self.scope = scope
        kwargs['typ'] = colander.Mapping(unknown='preserve')
        super(MetadataCollection, self).__init__(*args, **kwargs)

    def validator(self, node, cstruct):
        cstruct = cstruct or {}
        if len(cstruct) == 0:
            raise ValidationError(
                node,
                'At least one propety must be defined',
            )

        standards = METADATA_STANDARD_PROPERTIES.get(self.scope, {})

        for prop, value in iteritems(cstruct):
            if prop in standards:
                sub_schema(standards[prop], node, value)


class Value(colander.SchemaNode):
    value = colander.SchemaNode(
        AnyType(),
    )
    explanation = colander.SchemaNode(
        colander.String(),
        missing=colander.drop,
    )
    annotation = colander.SchemaNode(
        colander.String(),
        missing=colander.drop,
    )
    meta = MetadataCollection(
        METADATA_SCOPE_VALUE,
        missing=colander.drop,
    )

    def __init__(self, *args, **kwargs):
        kwargs['typ'] = colander.Mapping(unknown='raise')
        super(Value, self).__init__(*args, **kwargs)

    def validator(self, node, cstruct):
        if isinstance(cstruct['value'], list):
            for subtype in (
                    colander.SchemaNode(colander.String()),
                    ValueCollection):
                for value in cstruct['value']:
                    try:
                        sub_schema(subtype, node, value)
                    except ValidationError:
                        break
                else:
                    return

            raise ValidationError(
                node,
                'Lists must be consist only of Strings or ValueCollections',
            )

        elif isinstance(cstruct['value'], dict):
            sub_schema(ValueCollectionMapping, node, cstruct['value'])


class ValueCollection(colander.SchemaNode):
    def __init__(self, *args, **kwargs):
        kwargs['typ'] = colander.Mapping(unknown='preserve')
        super(ValueCollection, self).__init__(*args, **kwargs)

    def validator(self, node, cstruct):
        cstruct = cstruct or {}
        if len(cstruct) == 0:
            raise ValidationError(
                node,
                'At least one Value must be defined',
            )

        for field_id, value in iteritems(cstruct):
            sub_schema(IdentifierString, node, field_id)
            sub_schema(Value, node, value)


class ValueCollectionMapping(colander.SchemaNode):
    def __init__(self, *args, **kwargs):
        kwargs['typ'] = colander.Mapping(unknown='preserve')
        super(ValueCollectionMapping, self).__init__(*args, **kwargs)

    def validator(self, node, cstruct):
        cstruct = cstruct or {}
        if len(cstruct) == 0:
            raise ValidationError(
                node,
                'At least one Row must be defined',
            )

        for field_id, values in iteritems(cstruct):
            sub_schema(IdentifierString, node, field_id)
            sub_schema(ValueCollection, node, values)


VALUE_TYPE_CHECKS = {
    'integer': lambda val: isinstance(val, integer_types),
    'float': lambda val: isinstance(val, (float,) + integer_types),
    'text': lambda val: isinstance(val, string_types),
    'enumeration': lambda val: isinstance(val, string_types),
    'boolean': lambda val: isinstance(val, bool),
    'date': lambda val: isinstance(val, string_types) and RE_DATE.match(val),
    'time': lambda val: isinstance(val, string_types) and RE_TIME.match(val),
    'dateTime':
    lambda val: isinstance(val, string_types) and RE_DATETIME.match(val),
    'enumerationSet': lambda val: isinstance(val, list),
    'recordList': lambda val: isinstance(val, list),
    'matrix': lambda val: isinstance(val, dict),
}


class Assessment(colander.SchemaNode):
    instrument = InstrumentReference()
    meta = MetadataCollection(
        METADATA_SCOPE_ASSESSMENT,
        missing=colander.drop,
    )
    values = ValueCollection()

    def __init__(self, instrument=None, *args, **kwargs):
        self.instrument = instrument
        kwargs['typ'] = colander.Mapping(unknown='raise')
        super(Assessment, self).__init__(*args, **kwargs)

    def validator(self, node, cstruct):
        if not self.instrument:
            return

        validate_instrument_version(
            self.instrument,
            cstruct,
            node.get('instrument'),
        )

        self.check_has_all_fields(
            node.get('values'),
            cstruct['values'],
            self.instrument['record'],
        )

    def check_has_all_fields(self, node, values, fields):
        values = deepcopy(values)
        if not isinstance(values, dict):
            raise ValidationError(
                node,
                'Value expected to contain a mapping.'
            )

        for field in fields:
            value = values.pop(field['id'], None)
            if value is None:
                raise ValidationError(
                    node,
                    'No value exists for field ID "%s"' % field['id'],
                )

            if value['value'] is None and field.get('required', False):
                raise ValidationError(
                    node,
                    'No value present for required field ID "%s"' % (
                        field['id'],
                    ),
                )

            full_type_def = get_full_type_definition(
                self.instrument,
                field['type'],
            )

            self._check_value_type(node, value['value'], field, full_type_def)

            self._check_metafields(node, value, field)

            self._check_complex_subfields(node, full_type_def, value)

        if len(values) > 0:
            raise ValidationError(
                node,
                'Unknown field IDs found: %s' % ', '.join(list(values.keys())),
            )

    def _check_value_type(self, node, value, field, type_def):
        if value is None:
            return

        error = ValidationError(
            node,
            'Value for "%s" is not of the correct type' % (
                field['id'],
            ),
        )

        # Basic checks
        if not VALUE_TYPE_CHECKS[type_def['base']](value):
            raise error

        # Deeper checks
        if type_def['base'] == 'enumerationSet':
            for subval in value:
                if not isinstance(subval, string_types):
                    raise error

    def _check_metafields(self, node, value, field):
        explanation = field.get('explanation', 'none')
        if 'explanation' in value \
                and value['explanation'] is not None \
                and explanation == 'none':
            raise ValidationError(
                node,
                'Explanation present where not allowed in field ID "%s"' % (
                    field['id'],
                ),
            )
        elif 'explanation' not in value and explanation == 'required':
            raise ValidationError(
                node,
                'Explanation missing for field ID "%s"' % (
                    field['id'],
                ),
            )

        annotation = field.get('annotation', 'none')
        if 'annotation' in value and value['annotation'] is not None:
            if annotation == 'none':
                raise ValidationError(
                    node,
                    'Annotation present where not allowed',
                )

            elif value['value'] is not None:
                raise ValidationError(
                    node,
                    'Annotation provided for non-empty value',
                )
        elif 'annotation' not in value \
                and annotation == 'required' \
                and value['value'] is None:
            raise ValidationError(
                node,
                'Annotation missing for field ID "%s"' % (
                    field['id'],
                ),
            )

    def _check_complex_subfields(self, node, full_type_def, value):
        if value['value'] is None:
            return

        if 'record' in full_type_def:
            for rec in value['value']:
                self.check_has_all_fields(
                    node,
                    rec,
                    full_type_def['record'],
                )

        elif 'rows' in full_type_def:
            for row in full_type_def['rows']:
                row_value = value['value'].pop(row['id'], None)
                if row_value is None:
                    raise ValidationError(
                        node,
                        'Missing values for row ID "%s"' % row['id'],
                    )

                self.check_has_all_fields(
                    node,
                    row_value,
                    full_type_def['columns'],
                )

            if len(value['value']) > 0:
                raise ValidationError(
                    node,
                    'Unknown row IDs found: %s' % (
                        ', '.join(list(value['value'].keys())),
                    ),
                )




#
# Copyright (c) 2015, Prometheus Research, LLC
#


import json

import six

from .common import get_json, get_yaml
from .assessment import Assessment
from .calculationset import CalculationSet
from .form import Form
from .instrument import Instrument
from .interaction import Interaction


__all__ = (
    'get_instrument_json',
    'get_instrument_yaml',

    'get_assessment_json',
    'get_assessment_yaml',

    'get_form_json',
    'get_form_yaml',

    'get_calculationset_json',
    'get_calculationset_yaml',

    'get_interaction_json',
    'get_interaction_yaml',
)


def _get_struct(src):
    if isinstance(src, six.string_types):
        src = json.loads(src)
    elif hasattr(src, 'read'):
        src = json.load(src)
    return src


def get_instrument_json(instrument, pretty=True, **kwargs):
    """
    Generates a JSON-formatted string containing the specified Instrument
    Definition.

    :param instrument: The Instrument Definition generate the JSON for
    :type instrument: JSON string, dict, or file-like object
    :param pretty:
        Whether or not to format the JSON in a human-friendly way. If not
        specified, defaults to ``True``.
    :type pretty: bool
    :param kwargs:
        Any extra keyword arguments are passed to the underlying ``json.dumps``
        function.
    :returns: The JSON-formatted string representing the Instrument
    :rtype: string
    """

    instrument = _get_struct(instrument)
    kwargs['pretty'] = pretty
    return get_json(Instrument(instrument), **kwargs)


def get_instrument_yaml(instrument, pretty=True, **kwargs):
    """
    Generates a YAML-formatted string containing the specified Instrument
    Definition.

    :param instrument: The Instrument Definition generate the YAML for
    :type instrument: JSON string, dict, or file-like object
    :param pretty:
        Whether or not to format the YAML in a human-friendly way. If not
        specified, defaults to ``True``.
    :type pretty: bool
    :param kwargs:
        Any extra keyword arguments are passed to the underlying ``yaml.dump``
        function.
    :returns: The YAML-formatted string representing the Instrument
    :rtype: string
    """

    instrument = _get_struct(instrument)
    kwargs['pretty'] = pretty
    return get_yaml(Instrument(instrument), **kwargs)


def get_assessment_json(assessment, pretty=True, **kwargs):
    """
    Generates a JSON-formatted string containing the specified Assessment
    Document.

    :param instrument: The Assessment Document generate the JSON for
    :type instrument: JSON string, dict, or file-like object
    :param pretty:
        Whether or not to format the JSON in a human-friendly way. If not
        specified, defaults to ``True``.
    :type pretty: bool
    :param kwargs:
        Any extra keyword arguments are passed to the underlying ``json.dumps``
        function.
    :returns: The JSON-formatted string representing the Assessment
    :rtype: string
    """

    assessment = _get_struct(assessment)
    kwargs['pretty'] = pretty
    return get_json(Assessment(assessment), **kwargs)


def get_assessment_yaml(assessment, pretty=True, **kwargs):
    """
    Generates a YAML-formatted string containing the specified Assessment
    Document.

    :param instrument: The Assessment Document generate the YAML for
    :type instrument: JSON string, dict, or file-like object
    :param pretty:
        Whether or not to format the YAML in a human-friendly way. If not
        specified, defaults to ``True``.
    :type pretty: bool
    :param kwargs:
        Any extra keyword arguments are passed to the underlying ``yaml.dump``
        function.
    :returns: The YAML-formatted string representing the Assessment
    :rtype: string
    """

    assessment = _get_struct(assessment)
    kwargs['pretty'] = pretty
    return get_yaml(Assessment(assessment), **kwargs)


def get_form_json(form, pretty=True, **kwargs):
    """
    Generates a JSON-formatted string containing the specified Web Form
    Configuration.

    :param instrument: The Web Form Configuration generate the JSON for
    :type instrument: JSON string, dict, or file-like object
    :param pretty:
        Whether or not to format the JSON in a human-friendly way. If not
        specified, defaults to ``True``.
    :type pretty: bool
    :param kwargs:
        Any extra keyword arguments are passed to the underlying ``json.dumps``
        function.
    :returns: The JSON-formatted string representing the Form
    :rtype: string
    """

    form = _get_struct(form)
    kwargs['pretty'] = pretty
    return get_json(Form(form), **kwargs)


def get_form_yaml(form, pretty=True, **kwargs):
    """
    Generates a YAML-formatted string containing the specified Web Form
    Configuration.

    :param instrument: The Web Form Configuration generate the YAML for
    :type instrument: JSON string, dict, or file-like object
    :param pretty:
        Whether or not to format the YAML in a human-friendly way. If not
        specified, defaults to ``True``.
    :type pretty: bool
    :param kwargs:
        Any extra keyword arguments are passed to the underlying ``yaml.dump``
        function.
    :returns: The YAML-formatted string representing the Form
    :rtype: string
    """

    form = _get_struct(form)
    kwargs['pretty'] = pretty
    return get_yaml(Form(form), **kwargs)


def get_calculationset_json(calculationset, pretty=True, **kwargs):
    """
    Generates a JSON-formatted string containing the specified Calculation Set
    Definition.

    :param instrument: The Calculation Set Definition generate the JSON for
    :type instrument: JSON string, dict, or file-like object
    :param pretty:
        Whether or not to format the JSON in a human-friendly way. If not
        specified, defaults to ``True``.
    :type pretty: bool
    :param kwargs:
        Any extra keyword arguments are passed to the underlying ``json.dumps``
        function.
    :returns: The JSON-formatted string representing the Calculation Set
    :rtype: string
    """

    calculationset = _get_struct(calculationset)
    kwargs['pretty'] = pretty
    return get_json(CalculationSet(calculationset), **kwargs)


def get_calculationset_yaml(calculationset, pretty=True, **kwargs):
    """
    Generates a YAML-formatted string containing the specified Calculation Set
    Definition.

    :param instrument: The Calculation Set Definition generate the YAML for
    :type instrument: JSON string, dict, or file-like object
    :param pretty:
        Whether or not to format the YAML in a human-friendly way. If not
        specified, defaults to ``True``.
    :type pretty: bool
    :param kwargs:
        Any extra keyword arguments are passed to the underlying ``yaml.dump``
        function.
    :returns: The YAML-formatted string representing the Calculation Set
    :rtype: string
    """

    calculationset = _get_struct(calculationset)
    kwargs['pretty'] = pretty
    return get_yaml(CalculationSet(calculationset), **kwargs)


def get_interaction_json(interaction, pretty=True, **kwargs):
    """
    Generates a JSON-formatted string containing the specified SMS Interaction
    Configuration.

    :param instrument: The SMS Interaction Configuration generate the JSON for
    :type instrument: JSON string, dict, or file-like object
    :param pretty:
        Whether or not to format the JSON in a human-friendly way. If not
        specified, defaults to ``True``.
    :type pretty: bool
    :param kwargs:
        Any extra keyword arguments are passed to the underlying ``json.dumps``
        function.
    :returns: The JSON-formatted string representing the Form
    :rtype: string
    """

    interaction = _get_struct(interaction)
    kwargs['pretty'] = pretty
    return get_json(Interaction(interaction), **kwargs)


def get_interaction_yaml(interaction, pretty=True, **kwargs):
    """
    Generates a YAML-formatted string containing the specified SMS Interaction
    Configuration.

    :param instrument: The SMS Interaction Configuration generate the YAML for
    :type instrument: JSON string, dict, or file-like object
    :param pretty:
        Whether or not to format the YAML in a human-friendly way. If not
        specified, defaults to ``True``.
    :type pretty: bool
    :param kwargs:
        Any extra keyword arguments are passed to the underlying ``yaml.dump``
        function.
    :returns: The YAML-formatted string representing the Form
    :rtype: string
    """

    interaction = _get_struct(interaction)
    kwargs['pretty'] = pretty
    return get_yaml(Interaction(interaction), **kwargs)




#
# Copyright (c) 2015, Prometheus Research, LLC
#


import re

from copy import deepcopy

import colander

from six import iteritems, iterkeys, string_types

from .common import ValidationError, RE_IDENTIFIER, IdentifierString, \
    sub_schema, AnyType, OneOfType, StrictBooleanType, OptionalStringType


__all__ = (
    'TYPES_SIMPLE',
    'TYPES_COMPLEX',
    'TYPES_ALL',
    'CONSTRAINTS_ALL',
    'TYPES_CONSTRAINED',
    'TYPES_CONSTRAINED_REQUIRED',
    'RE_ENUMERATION_ID',

    'get_full_type_definition',

    'InstrumentIdentifier',
    'InstrumentReference',
    'Version',
    'Description',
    'EnumerationIdentifier',
    'Enumeration',
    'EnumerationCollection',
    'BoundConstraint',
    'IntegerBoundConstraint',
    'Column',
    'ColumnCollection',
    'Row',
    'RowCollection',
    'TypeDefinition',
    'RequiredOptionalField',
    'InstrumentTypes',
    'FieldType',
    'Field',
    'Record',
    'Instrument',
)


TYPES_SIMPLE = (
    'text',
    'integer',
    'float',
    'boolean',
    'enumeration',
    'enumerationSet',
    'date',
    'time',
    'dateTime',
)

TYPES_COMPLEX = (
    'recordList',
    'matrix',
)

TYPES_ALL = TYPES_SIMPLE + TYPES_COMPLEX

CONSTRAINTS_ALL = (
    'range',
    'length',
    'pattern',
    'enumerations',
    'record',
    'columns',
    'rows',
)

TYPES_CONSTRAINED = {
    'integer': [
        'range',
    ],

    'float': [
        'range',
    ],

    'date': [
        'range',
    ],

    'time': [
        'range',
    ],

    'dateTime': [
        'range',
    ],

    'text': [
        'length',
        'pattern',
    ],

    'enumeration': [
        'enumerations',
    ],

    'enumerationSet': [
        'length',
        'enumerations',
    ],

    'recordList': [
        'length',
        'record',
    ],

    'matrix': [
        'rows',
        'columns',
    ],
}

TYPES_CONSTRAINED_REQUIRED = {
    'enumeration': [
        'enumerations',
    ],

    'enumerationSet': [
        'enumerations',
    ],

    'recordList': [
        'record',
    ],

    'matrix': [
        'rows',
        'columns',
    ],
}

RANGE_CONSTRAINT_TYPES = {
    'integer': colander.Integer(),
    'float': colander.Float(),
    'date': colander.Date(),
    'time': colander.Time(),
    'dateTime': colander.DateTime(),
}


RE_VERSION = re.compile(r'(0|[1-9][0-9]*)\.(0|[1-9][0-9]*)')
RE_ENUMERATION_ID = re.compile(
    r'^(?:[a-z0-9]{1,2}|[a-z0-9](?:[a-z0-9]|[_-](?![_-]))+[a-z0-9])$'
)


# pylint: disable=abstract-method


class InstrumentIdentifier(colander.SchemaNode):
    schema_type = colander.String
    validator = colander.url


class Version(colander.SchemaNode):
    schema_type = colander.String
    validator = colander.Regex(RE_VERSION)


class InstrumentReference(colander.SchemaNode):
    id = InstrumentIdentifier()  # pylint: disable=invalid-name
    version = Version()

    def __init__(self, *args, **kwargs):
        kwargs['typ'] = colander.Mapping(unknown='raise')
        super(InstrumentReference, self).__init__(*args, **kwargs)


class Description(colander.SchemaNode):
    schema_type = OptionalStringType
    validator = colander.Length(min=1)
    missing = colander.drop


class Enumeration(colander.MappingSchema):
    description = Description()


class EnumerationIdentifier(colander.SchemaNode):
    schema_type = colander.String
    validator = colander.Regex(RE_ENUMERATION_ID)


class EnumerationCollection(colander.SchemaNode):
    def __init__(self, *args, **kwargs):
        kwargs['typ'] = colander.Mapping(unknown='preserve')
        super(EnumerationCollection, self).__init__(*args, **kwargs)

    def validator(self, node, cstruct):
        cstruct = cstruct or {}
        if len(cstruct) == 0:
            raise ValidationError(
                node,
                'At least one Enumeration must be defined',
            )

        for enum_id, enum_def in iteritems(cstruct):
            sub_schema(EnumerationIdentifier, node, enum_id)
            if enum_def is not None:
                sub_schema(Enumeration, node, enum_def)


class BoundConstraint(colander.SchemaNode):
    def __init__(self, schema_type=None, **kwargs):
        self.schema_type = schema_type
        schema_type = schema_type or AnyType()
        super(BoundConstraint, self).__init__(
            colander.Mapping(unknown='raise'),
            colander.SchemaNode(
                schema_type,
                name='min',
                missing=colander.drop,
            ),
            colander.SchemaNode(
                schema_type,
                name='max',
                missing=colander.drop,
            ),
            **kwargs
        )

    def validator(self, node, cstruct):
        if len(cstruct) < 1:
            raise ValidationError(
                node,
                'At least one bound must be specified',
            )

        if self.schema_type:
            min_value = cstruct.get('min', None)
            max_value = cstruct.get('max', None)
            if min_value is not None \
                    and max_value is not None \
                    and min_value > max_value:
                raise ValidationError(
                    node,
                    'The minimum bound must be lower than than maximum',
                )


class IntegerBoundConstraint(BoundConstraint):
    def __init__(self, *args, **kwargs):
        super(IntegerBoundConstraint, self).__init__(
            *args,
            schema_type=colander.Integer(),
            **kwargs
        )


class FieldType(colander.SchemaNode):
    def __init__(self, *args, **kwargs):
        kwargs['typ'] = OneOfType(
            colander.String,
            colander.Mapping(unknown='preserve'),
        )
        super(FieldType, self).__init__(*args, **kwargs)

    def validator(self, node, cstruct):
        if isinstance(cstruct, string_types):
            if cstruct not in TYPES_ALL \
                    and not RE_IDENTIFIER.match(cstruct):
                raise ValidationError(
                    node,
                    '"%r" is not a valid type identifier' % (cstruct,),
                )
        else:
            sub_schema(TypeDefinition, node, cstruct)


class Column(colander.SchemaNode):
    id = IdentifierString()  # pylint: disable=invalid-name
    description = Description()
    required = colander.SchemaNode(
        StrictBooleanType(),
        missing=colander.drop,
    )
    identifiable = colander.SchemaNode(
        StrictBooleanType(),
        missing=colander.drop,
    )

    def __init__(self, *args, **kwargs):
        kwargs['typ'] = colander.Mapping(unknown='raise')
        super(Column, self).__init__(*args, **kwargs)
        self.add(FieldType(name='type'))


class ColumnCollection(colander.SequenceSchema):
    column = Column()

    def validator(self, node, cstruct):
        if len(cstruct) < 1:
            raise ValidationError(
                node,
                'Shorter than minimum length 1',
            )

        ids = [col['id'] for col in cstruct]
        if len(ids) != len(set(ids)):
            raise ValidationError(
                node,
                'Column IDs must be unique within a collection',
            )


class Row(colander.SchemaNode):
    id = IdentifierString()  # pylint: disable=invalid-name
    description = Description()
    required = colander.SchemaNode(
        StrictBooleanType(),
        missing=colander.drop,
    )

    def __init__(self, *args, **kwargs):
        kwargs['typ'] = colander.Mapping(unknown='raise')
        super(Row, self).__init__(*args, **kwargs)


class RowCollection(colander.SequenceSchema):
    row = Row()

    def validator(self, node, cstruct):
        if len(cstruct) < 1:
            raise ValidationError(
                node,
                'Shorter than minimum length 1',
            )

        ids = [row['id'] for row in cstruct]
        if len(ids) != len(set(ids)):
            raise ValidationError(
                node,
                'Row IDs must be unique within a collection',
            )


class RangeConstraint(BoundConstraint):
    def __init__(self, *args, **kwargs):
        super(RangeConstraint, self).__init__(
            *args,
            **kwargs
        )


class TypeDefinition(colander.SchemaNode):
    base = colander.SchemaNode(colander.String())
    range = RangeConstraint(missing=colander.drop)
    length = IntegerBoundConstraint(missing=colander.drop)
    pattern = colander.SchemaNode(
        colander.String(),
        missing=colander.drop,
    )
    enumerations = EnumerationCollection(missing=colander.drop)
    columns = ColumnCollection(missing=colander.drop)
    rows = RowCollection(missing=colander.drop)

    def __init__(self, *args, **kwargs):
        kwargs['typ'] = colander.Mapping(unknown='raise')
        super(TypeDefinition, self).__init__(*args, **kwargs)
        self.add(Record(name='record', missing=colander.drop))


class RequiredOptionalField(colander.SchemaNode):
    schema_type = colander.String
    validator = colander.OneOf([
        'required',
        'optional',
        'none',
    ])
    missing = colander.drop


class Field(colander.SchemaNode):
    id = IdentifierString()  # pylint: disable=invalid-name
    description = Description()
    type = FieldType()
    required = colander.SchemaNode(
        StrictBooleanType(),
        missing=colander.drop,
    )
    identifiable = colander.SchemaNode(
        StrictBooleanType(),
        missing=colander.drop,
    )
    annotation = RequiredOptionalField()
    explanation = RequiredOptionalField()

    def __init__(self, *args, **kwargs):
        kwargs['typ'] = colander.Mapping(unknown='raise')
        super(Field, self).__init__(*args, **kwargs)

    def validator(self, node, cstruct):
        if 'annotation' in cstruct and cstruct['annotation'] != 'none':
            if 'required' in cstruct and cstruct['required']:
                raise ValidationError(
                    node,
                    'A Field cannot have an annotation if it is required',
                )


class Record(colander.SequenceSchema):
    field = Field()

    def validator(self, node, cstruct):
        if len(cstruct) < 1:
            raise ValidationError(
                node,
                'Shorter than minimum length 1',
            )

        ids = [field['id'] for field in cstruct]
        if len(ids) != len(set(ids)):
            raise ValidationError(
                node,
                'Field IDs must be unique within a record',
            )


class InstrumentTypes(colander.SchemaNode):
    def __init__(self, *args, **kwargs):
        kwargs['typ'] = colander.Mapping(unknown='preserve')
        super(InstrumentTypes, self).__init__(*args, **kwargs)

    def validator(self, node, cstruct):
        cstruct = cstruct or {}
        for type_id, type_def in iteritems(cstruct):
            if type_id in TYPES_ALL or not RE_IDENTIFIER.match(type_id):
                raise ValidationError(
                    node,
                    '"%r" is not a valid custom type ID' % type_id,
                )
            sub_schema(TypeDefinition, node, type_def)


class Instrument(colander.SchemaNode):
    id = InstrumentIdentifier()  # pylint: disable=invalid-name
    version = Version()
    title = colander.SchemaNode(colander.String())
    description = Description()
    types = InstrumentTypes(missing=colander.drop)
    record = Record()

    def __init__(self, *args, **kwargs):
        kwargs['typ'] = colander.Mapping(unknown='raise')
        super(Instrument, self).__init__(*args, **kwargs)

    def validator(self, node, cstruct):
        for _, type_def in iteritems(cstruct.get('types', {})):
            self.check_type(type_def, node.get('types'), cstruct)
        for field in cstruct['record']:
            self.check_type(field['type'], node.get('record'), cstruct)

    def check_type(self, type_def, node, cstruct):
        try:
            full_type_def = get_full_type_definition(cstruct, type_def)
        except Exception as exc:
            raise ValidationError(
                node,
                str(exc),
            )

        self._check_required_constraints(full_type_def, node, type_def)
        self._check_appropriate_constraints(full_type_def, node)
        self._check_range_constraints(full_type_def, node)
        self._check_complex_subfields(full_type_def, node, cstruct)

        return full_type_def

    def _check_required_constraints(self, full_type_def, node, cstruct):
        if full_type_def['base'] in iterkeys(TYPES_CONSTRAINED_REQUIRED):
            for con in TYPES_CONSTRAINED_REQUIRED[full_type_def['base']]:
                if con not in full_type_def:
                    raise ValidationError(
                        node,
                        'Type definition "%r" missing required constraint'
                        ' "%s"' % (
                            cstruct,
                            con,
                        ),
                    )

    def _check_appropriate_constraints(self, full_type_def, node):
        for con in CONSTRAINTS_ALL:
            if con in full_type_def:
                if con not in TYPES_CONSTRAINED.get(full_type_def['base'], []):
                    raise ValidationError(
                        node,
                        'Constraint "%s" cannot be used on types based on'
                        ' "%s"' % (
                            con,
                            full_type_def['base'],
                        ),
                    )

    def _check_range_constraints(self, full_type_def, node):
        if 'range' in full_type_def \
                and full_type_def['base'] in RANGE_CONSTRAINT_TYPES:
            sub_schema(
                BoundConstraint(RANGE_CONSTRAINT_TYPES[full_type_def['base']]),
                node,
                full_type_def['range'],
            )

    def _check_complex_subfields(self, full_type_def, node, instrument):
        for sub_field_constraint in ('record', 'columns'):
            if sub_field_constraint in full_type_def:
                for field in full_type_def[sub_field_constraint]:
                    sub_type = self.check_type(field['type'], node, instrument)
                    if sub_type['base'] in TYPES_COMPLEX:
                        raise ValidationError(
                            node,
                            'Complex types cannot contain other complex'
                            ' types.',
                        )


def get_full_type_definition(instrument, type_def):
    if isinstance(type_def, string_types):
        if type_def in TYPES_ALL:
            return {
                'base': type_def
            }

        if type_def in iterkeys(instrument.get('types', {})):
            return get_full_type_definition(
                instrument,
                instrument['types'][type_def],
            )

        raise ValueError(
            'no type is defined for identifier "%s"' % (
                type_def,
            )
        )

    elif isinstance(type_def, dict):
        type_def = deepcopy(type_def)
        base_type = type_def.pop('base')

        try:
            parent_type_def = get_full_type_definition(instrument, base_type)
        except ValueError:
            raise ValueError(
                'invalid definition, references undefined base type "%s"' % (
                    base_type,
                )
            )

        parent_type_def.update(type_def)
        return parent_type_def

    else:
        raise TypeError('type_def must be a string or dict, got "%r"')




#
# Copyright (c) 2015, Prometheus Research, LLC
#


from setuptools import setup, find_packages


setup(
    name='prismh.core',
    version='0.1.0',
    description='Parsing and Validation library for PRISMH Files',
    long_description=open('README.rst', 'r').read(),
    keywords='prismh instrument assessment validation',
    author='Prometheus Research, LLC',
    author_email='contact@prometheusresearch.com',
    license='AGPLv3',
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU Affero General Public License v3',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
    ],
    url='https://bitbucket.org/prometheus/prismh.core',
    package_dir={'': 'src'},
    packages=find_packages('src'),
    zip_safe=True,
    include_package_data=True,
    namespace_packages=['prismh'],
    install_requires=[
        'six>=1.8,<2',
        'colander>=1.0,<1.1',
        'pyyaml',
    ],
    extras_require={
        'dev': [
            'coverage>=3.7,<4',
            'nose>=1.3,<2',
            'nosy>=1.1,<2',
            'prospector[with_pyroma]>=0.10,<0.11',
            'twine>=1.5,<2',
            'wheel>=0.24,<0.25',
            'Sphinx>=1.3,<2',
            'sphinx-autobuild>=0.5,<0.6',
            'tox>=2,<3',
        ],
    },
    test_suite='nose.collector',
)




#
# Copyright (c) 2015, Prometheus Research, LLC
#


from .common import SortedDict, TypedSortedDict, TypedDefinedOrderDict, \
    InstrumentReference, Descriptor


__all__ = (
    'Form',
)


class Event(TypedDefinedOrderDict):
    order = [
        'trigger',
        'action',
        'targets',
        'options',
    ]

    key_types = {
        'options': SortedDict,
    }


class Widget(TypedDefinedOrderDict):
    order = [
        'type',
        'options',
    ]

    key_types = {
        'options': SortedDict,
    }


class ElementOptions(TypedDefinedOrderDict):
    order = [
        'fieldId',
        'text',
        'help',
        'error',
        'audio',
        'enumerations',
        'questions',
        'rows',
        'widget',
        'events',
    ]

    key_types = {
        'text': SortedDict,
        'help': SortedDict,
        'error': SortedDict,
        'audio': SortedDict,
        'enumerations': [Descriptor],
        'rows': [Descriptor],
        'widget': Widget,
        'events': [Event],
    }

ElementOptions.key_types['questions'] = [ElementOptions]


class Element(TypedDefinedOrderDict):
    order = [
        'type',
        'tags',
        'options',
    ]

    key_types = {
        'options': ElementOptions,
    }


class Page(TypedDefinedOrderDict):
    order = [
        'id',
        'elements',
    ]

    key_types = {
        'elements': [Element],
    }


class Unprompted(TypedDefinedOrderDict):
    order = [
        'action',
        'options',
    ]

    key_types = {
        'options': SortedDict,
    }


class UnpromptedCollection(TypedSortedDict):
    subtype = Unprompted


class Parameter(TypedDefinedOrderDict):
    order = [
        'type',
    ]


class ParameterCollection(TypedSortedDict):
    subtype = Parameter


class Form(TypedDefinedOrderDict):
    order = [
        'instrument',
        'defaultLocalization',
        'title',
        'pages',
        'unprompted',
        'parameters',
    ]

    key_types = {
        'instrument': InstrumentReference,
        'title': SortedDict,
        'pages': [Page],
        'unprompted': UnpromptedCollection,
        'parameters': ParameterCollection,
    }




#
# Copyright (c) 2015, Prometheus Research, LLC
#


import collections
import json

import yaml

from six import text_type, add_metaclass


__all__ = (
    'get_json',
    'get_yaml',

    'OrderedDict',
    'SortedDict',
    'TypedSortedDict',
    'DefinedOrderDict',
    'TypedDefinedOrderDict',

    'InstrumentReference',
    'Descriptor',
)


def get_json(data, pretty=False, **kwargs):
    """
    A convenience wrapper around ``json.dumps`` that respects the ordering of
    keys in classes like ``OrderedDict``, ``SortedDict``, and
    ``DefinedOrderDict``.

    :param data: the object to encode in JSON
    :param pretty:
        whether or not the output should be indented in human-friendly ways
    :type pretty: boolean
    :returns: a JSON-encoded string
    """

    kwargs['ensure_ascii'] = False
    kwargs['sort_keys'] = False

    if pretty:
        kwargs['indent'] = 2
        kwargs['separators'] = (',', ': ')

    return json.dumps(data, **kwargs)


def get_yaml(data, pretty=False, **kwargs):
    """
    A convenience wrapper around ``yaml.dump`` that respects the ordering of
    keys in classes like ``OrderedDict``, ``SortedDict``, and
    ``DefinedOrderDict``.

    :param data: the object to encode in YAML
    :param pretty:
        whether or not the output should be indented in human-friendly ways
    :type pretty: boolean
    :returns: a YAML-encoded string
    """

    kwargs['Dumper'] = OrderedDumper
    kwargs['allow_unicode'] = True

    if pretty:
        kwargs['default_flow_style'] = False

    return yaml.dump(data, **kwargs).rstrip()


class OrderedDumper(yaml.Dumper):
    pass


def unicode_representer(dumper, ustr):
    return dumper.represent_scalar(
        'tag:yaml.org,2002:str',
        ustr,
    )


OrderedDumper.add_representer(text_type, unicode_representer)


def dict_representer(dumper, data):
    return dumper.represent_mapping(
        yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG,
        list(data.items())
    )


class OrderedDumperMetaclass(type):
    def __init__(cls, name, bases, dct):  # noqa
        super(OrderedDumperMetaclass, cls).__init__(name, bases, dct)
        OrderedDumper.add_representer(cls, dict_representer)


@add_metaclass(OrderedDumperMetaclass)
class OrderedDict(collections.OrderedDict):
    """
    A functional equivalent to ``collections.OrderedDict``.
    """

    pass


@add_metaclass(OrderedDumperMetaclass)
class SortedDict(dict):
    """
    A dictionary class that sorts its keys alphabetically.
    """

    def get_keys(self):
        return sorted(super(SortedDict, self).keys())

    def keys(self):
        return self.get_keys()

    def __iter__(self):
        return iter(self.get_keys())

    def items(self):
        return [
            (key, self[key])
            for key in self.get_keys()
        ]

    def iteritems(self):
        for key in self.get_keys():
            yield (key, self[key])


class TypedSortedDict(SortedDict):
    """
    A variety of the ``SortedDict`` class that automatically casts the value of
    all keys to the type specified on the ``subtype`` property.
    """

    #: The type to cast all values in the dictionary to.
    subtype = None

    def __init__(self, obj=None):
        super(TypedSortedDict, self).__init__(obj or {})
        for key in self:
            self[key] = self[key]

    def __setitem__(self, key, value):
        if self.subtype:
            # pylint: disable=E1102
            value = self.subtype(value)
        super(TypedSortedDict, self).__setitem__(key, value)


class DefinedOrderDict(SortedDict):
    """
    A dictionary class that orders its keys according to its ``order``
    property (any unnamed keys are then sorted alphabetically).
    """

    #: A list of key names in the order you want them to be output.
    order = []

    def get_keys(self):
        existing_keys = super(DefinedOrderDict, self).get_keys()
        keys = []
        for key in self.order:
            if key in existing_keys:
                keys.append(key)
        for key in existing_keys:
            if key not in self.order:
                keys.append(key)
        return keys


class TypedDefinedOrderDict(DefinedOrderDict):
    """
    A variety of the ``DefinedOrderDict`` class that provides for the automatic
    casting of values in the dictionary based on their key. This conversion is
    driven by the ``key_types`` property. E.g.::

        key_types = {
            'foo': SortedOrderDict,
            'bar': [SortedOrderDict],
        }
    """

    #: The mapping of key names to types. To indicate that the key should
    #: contain a list of casted values, place the type in a list with one
    # element.
    key_types = {}

    def __init__(self, obj=None):
        super(TypedDefinedOrderDict, self).__init__(obj or {})
        for key in self.key_types:
            if key in self:
                # Force an invokation of our __setitem__
                self[key] = self[key]

    def __setitem__(self, key, value):
        if key in self.key_types:
            type_ = self.key_types[key]
            if isinstance(type_, list):
                value = [
                    type_[0](val)
                    for val in value
                ]
            else:
                value = type_(value)
        super(TypedDefinedOrderDict, self).__setitem__(key, value)


class InstrumentReference(TypedDefinedOrderDict):
    order = [
        'id',
        'version',
    ]

    key_types = {
        'version': str,
    }


class Descriptor(TypedDefinedOrderDict):
    order = [
        'id',
        'text',
        'help',
        'audio',
    ]

    key_types = {
        'id': str,
        'text': SortedDict,
        'help': SortedDict,
        'audio': SortedDict,
    }

